#![cfg(feature = "server")]
use std::{net::SocketAddr, time::Duration};

use anyhow::{anyhow, Result};
use libpt::{
    bintols::display::humanbytes,
    log::{debug, error, info, trace},
};
use threadpool::ThreadPool;
use tokio::{
    io::{AsyncReadExt, BufReader},
    net::{TcpListener, TcpStream},
    time::timeout,
};

use crate::common::conf::Config;

pub mod listener;

pub struct Server {
    cfg: Config,
    pool: ThreadPool,
    pub timeout: Option<Duration>,
    server: TcpListener,
}

impl Server {
    pub async fn build(cfg: Config) -> Result<Self> {
        let server = TcpListener::bind(cfg.addr).await?;
        let pool = ThreadPool::new(cfg.threads);
        let timeout = Some(Duration::from_secs(5));
        Ok(Server {
            cfg,
            pool,
            timeout,
            server,
        })
    }
    pub async fn run(self) -> Result<()> {
        loop {
            let stream = self.server.accept().await?;
            self.handle_stream(stream).await?;
        }
    }

    async fn handle_stream(&self, stream: (TcpStream, SocketAddr)) -> Result<()> {
        const BUF_SIZE: usize = 1024;
        info!("start handling stream {:?}", stream.1);
        let mut buf = [0u8; BUF_SIZE];
        let mut reader = BufReader::new(stream.0);
        let mut len;
        loop {
            trace!("reading anew");
            len = self.read(&mut reader, &mut buf).await?;
            trace!("did read");
            if len == 0 {
                break;
            } else {
                info!(
                    "< {}\t({})\n{}",
                    stream.1,
                    humanbytes(len),
                    self.decode(&buf)?
                );
                buf = [0; BUF_SIZE];
            }
        }
        info!("stop handling stream {:?}", stream.1);
        Ok(())
    }

    #[inline]
    fn decode(&self, buf: &[u8]) -> Result<String> {
        Ok(std::str::from_utf8(buf)?.to_string())
    }

    async fn read(&self, reader: &mut BufReader<TcpStream>, buf: &mut [u8]) -> Result<usize> {
        let mut len: usize;
        let mut failsafe: u8 = 0;
        while failsafe < 5 {
            trace!("loop");
            len = match timeout(self.cfg.timeout, reader.read(buf)).await {
                Ok(inner) => {
                    trace!("Read for len: {inner:?}");
                    0
                }
                Err(err) => {
                    error!("timeout while reading: {err}");
                    return Err(err.into());
                }
            };
            if self.should_end(&buf) {
                return Ok(len);
            } else {
                // ?????????
                failsafe += 1;
            }
        }
        Err(anyhow!("read too often, so the failsafe activated"))
    }

    fn should_end(&self, buf: &[u8]) -> bool {
        let mut lb: u8 = buf[0];
        let mut iter = buf.iter().skip(1).peekable();
        while let Some(b) = iter.next() {
            if lb != 0 && *b == 0 && **iter.peek().unwrap() == 0 {
                return true
            }
            lb = *b;
        }
        return false
    }
}
